@@ -1,5 +1,7 @@
 /* GNU dump extensions to tar.
-   Copyright 1988,92,93,94,96,97,99,2000, 2001 Free Software Foundation, Inc.
+
+   Copyright (C) 1988, 1992, 1993, 1994, 1996, 1997, 1999, 2000, 2001,
+   2003 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -87,8 +89,8 @@ struct directory
     dev_t device_number;	/* device number for directory */
     ino_t inode_number;		/* inode number for directory */
     enum children children;
-    char nfs;
-    char found;
+    bool nfs;
+    bool found;
     char name[1];		/* path name of directory */
   };
 
@@ -126,7 +128,8 @@ compare_directories (void const *entry1, void const *entry2)
 static struct directory *
 note_directory (char const *name, dev_t dev, ino_t ino, bool nfs, bool found)
 {
-  struct directory *directory = xmalloc (sizeof *directory + strlen (name));
+  size_t size = offsetof (struct directory, name) + strlen (name) + 1;
+  struct directory *directory = xmalloc (size);
 
   directory->device_number = dev;
   directory->inode_number = ino;
@@ -152,7 +155,8 @@ find_directory (char *name)
     return 0;
   else
     {
-      struct directory *dir = alloca (sizeof *dir + strlen (name));
+      size_t size = offsetof (struct directory, name) + strlen (name) + 1;
+      struct directory *dir = alloca (size);
       strcpy (dir->name, name);
       return hash_lookup (directory_table, dir);
     }
@@ -183,7 +187,9 @@ get_directory_contents (char *path, dev_t device)
     enum children children;
 
     if (! dirp)
-      savedir_error (path);
+      {
+        savedir_error (path);
+      }
     errno = 0;
 
     name_buffer_size = strlen (path) + NAME_FIELD_SIZE;
@@ -198,7 +204,7 @@ get_directory_contents (char *path, dev_t device)
 
     accumulator = new_accumulator ();
 
-    if (children != NO_CHILDREN)
+    if (dirp && children != NO_CHILDREN)
       for (entry = dirp;
 	   (entrylen = strlen (entry)) != 0;
 	   entry += entrylen + 1)
@@ -220,7 +226,7 @@ get_directory_contents (char *path, dev_t device)
 
 	      if (deref_stat (dereference_option, name_buffer, &stat_data))
 		{
-		  stat_error (name_buffer);
+		  stat_diag (name_buffer);
 		  continue;
 		}
 
@@ -259,10 +265,10 @@ get_directory_contents (char *path, dev_t device)
 		      directory = note_directory (name_buffer,
 						  stat_data.st_dev,
 						  stat_data.st_ino, nfs, 1);
-		      directory->children = 
+		      directory->children =
 			((listed_incremental_option
 			  || newer_mtime_option <= stat_data.st_mtime
-			  || (after_date_option && 
+			  || (after_date_option &&
 			      newer_ctime_option <= stat_data.st_ctime))
 			 ? ALL_CHILDREN
 			 : CHANGED_CHILDREN);
@@ -305,7 +311,8 @@ get_directory_contents (char *path, dev_t device)
     add_to_accumulator (accumulator, "\000\000", 2);
 
     free (name_buffer);
-    free (dirp);
+    if (dirp)
+      free (dirp);
   }
 
   /* Sort the contents of the directory, now that we have it all.  */
@@ -405,7 +412,7 @@ read_directory_file (void)
 	{
 	  dev_t dev;
 	  ino_t ino;
-	  int nfs = buf[0] == '+';
+	  bool nfs = buf[0] == '+';
 	  char *strp = buf + nfs;
 
 	  lineno++;
@@ -499,7 +506,7 @@ write_directory_file (void)
 /* Restoration of incremental dumps.  */
 
 void
-gnu_restore (size_t skipcrud)
+gnu_restore (char const *directory_name)
 {
   char *archive_dir;
   char *current_dir;
@@ -509,9 +516,7 @@ gnu_restore (size_t skipcrud)
   union block *data_block;
   char *to;
 
-#define CURRENT_FILE_NAME (skipcrud + current_file_name)
-
-  current_dir = savedir (CURRENT_FILE_NAME);
+  current_dir = savedir (directory_name);
 
   if (!current_dir)
     {
@@ -522,8 +527,8 @@ gnu_restore (size_t skipcrud)
       return;
     }
 
-  size = current_stat.st_size;
-  if (size != current_stat.st_size)
+  size = current_stat_info.stat.st_size;
+  if (size != current_stat_info.stat.st_size)
     xalloc_die ();
   archive_dir = xmalloc (size);
   to = archive_dir;
@@ -554,7 +559,7 @@ gnu_restore (size_t skipcrud)
 	}
       if (*arc == '\0')
 	{
-	  char *p = new_name (CURRENT_FILE_NAME, cur);
+	  char *p = new_name (directory_name, cur);
 	  if (! interactive_option || confirm ("delete", p))
 	    {
 	      if (verbose_option)
@@ -572,6 +577,4 @@ gnu_restore (size_t skipcrud)
     }
   free (current_dir);
   free (archive_dir);
-
-#undef CURRENT_FILE_NAME
 }
